import math as m

from OSMPythonTools.api import Api
from OSMPythonTools.overpass import Overpass, overpassQueryBuilder, Nominatim
from structs.landmarks import Landmark, LandmarkType
from structs.preferences import Preferences, Preference
from typing import List
from typing import Tuple


BBOX_SIDE = 10              # size of bbox in *km* for general area, 10km
RADIUS_CLOSE_TO = 25        # size of area in *m* for close features, 25m radius
MIN_SCORE = 30              # DEPRECIATED. discard elements with score < 30
MIN_TAGS = 5                # DEPRECIATED. discard elements withs less than 5 tags
CHURCH_PENALTY = 0.1        # penalty to reduce score of curches
PARK_COEFF = 10             # multiplier for parks
N_IMPORTANT = 30            # take the 30 most important landmarks

SIGHTSEEING = LandmarkType(landmark_type='sightseeing')
NATURE = LandmarkType(landmark_type='nature')
SHOPPING = LandmarkType(landmark_type='shopping')


# Include the json here
# Create a list of all things to visit given some preferences and a city. Ready for the optimizer
def generate_landmarks(preferences: Preferences, city_country: str = None, coordinates: Tuple[float, float] = None)->Tuple[List[Landmark], List[Landmark]] :

    l_sights = ["'tourism'='museum'", "'tourism'='attraction'", "'tourism'='gallery'", 'historic', "'amenity'='arts_centre'", "'amenity'='planetarium'", "'amenity'='place_of_worship'", "'amenity'='fountain'", '"water"="reflecting_pool"'] 
    l_nature = ["'leisure'='park'", 'geological', "'natural'='geyser'", "'natural'='hot_spring'", '"natural"="arch"', '"natural"="cave_entrance"', '"natural"="volcano"', '"natural"="stone"', '"tourism"="alpine_hut"', '"tourism"="picnic_site"', '"tourism"="viewpoint"', '"tourism"="zoo"', '"waterway"="waterfall"'] 
    l_shop = ["'shop'='department_store'", "'shop'='mall'"] #, '"shop"="collector"', '"shop"="antiques"'] 
  
    L = []

    # Use 'City, Country'
    if city_country is not None :

        # List for sightseeing
        if preferences.sightseeing.score != 0 :
            L1 = get_landmarks_nominatim(city_country, l_sights, SIGHTSEEING)
            correct_score(L1, preferences.sightseeing)
            L += L1
        
        # List for nature
        if preferences.nature.score != 0 :
            L2 = get_landmarks_nominatim(city_country, l_nature, NATURE)
            correct_score(L2, preferences.nature)
            L += L2
        
        # List for shopping
        if preferences.shopping.score != 0 :
            L3 = get_landmarks_nominatim(city_country, l_shop, SHOPPING)
            correct_score(L3, preferences.shopping)
            L += L3

    # Use coordinates
    elif coordinates is not None :

        # List for sightseeing
        if preferences.sightseeing.score != 0 :
            L1 = get_landmarks_coords(coordinates, l_sights, SIGHTSEEING)
            correct_score(L1, preferences.sightseeing)
            L += L1
        
        # List for nature
        if preferences.nature.score != 0 :
            L2 = get_landmarks_coords(coordinates, l_nature, NATURE)
            correct_score(L2, preferences.nature)
            L += L2
        
        # List for shopping
        if preferences.shopping.score != 0 :
            L3 = get_landmarks_coords(coordinates, l_shop, SHOPPING)
            correct_score(L3, preferences.shopping)
            L += L3


    return L, take_most_important(L)
    #return L, cleanup_list(L)

# Determines if two locations are close to each other
def is_close_to(loc1: Tuple[float, float], loc2: Tuple[float, float])->bool :

    alpha = (180*RADIUS_CLOSE_TO)/(6371000*m.pi)
    if abs(loc1[0] - loc2[0]) + abs(loc1[1] - loc2[1]) < alpha*2 :
        return True
    else : 
        return False


def take_most_important(L: List[Landmark])->List[Landmark] :
    L_copy = []
    L_clean = []
    scores = [0]*len(L)
    names = []
    name_id = {}

    for i, elem in enumerate(L) :
        if elem.name not in names :
            names.append(elem.name)
            name_id[elem.name] = [i]
            L_copy.append(elem)
        else :
            name_id[elem.name] += [i]
            scores = []
            for j in name_id[elem.name] :
                scores.append(L[j].attractiveness)
            best_id = max(range(len(scores)), key=scores.__getitem__)
            t = name_id[elem.name][best_id]
            if t == i :
                for old in L_copy :
                    if old.name == elem.name :
                        old.attractiveness = L[t].attractiveness

            continue

    for i, elem in enumerate(L_copy) :
        scores[i] = elem.attractiveness

    res = sorted(range(len(scores)), key = lambda sub: scores[sub])[-N_IMPORTANT:]

    for i, elem in enumerate(L_copy) :
        if i in res :
            L_clean.append(elem)

    return L_clean


# Remove duplicate elements and elements with low score
def cleanup_list(L: List[Landmark])->List[Landmark] :
    L_clean = []
    names = []

    for landmark in L :

        if landmark.name in names :                 # Remove duplicates
            continue     

        elif landmark.attractiveness < MIN_SCORE :  # Remove uninteresting
            continue

        elif landmark.n_tags < MIN_TAGS :           # Remove uninteresting 2.0
            continue

        else :
            names.append(landmark.name)
            L_clean.append(landmark)
    
    return L_clean

# Correct the score of a list of landmarks by taking into account preferences and the number of tags
def correct_score(L: List[Landmark], preference: Preference) :

    if len(L) == 0 :
        return
    
    if L[0].type != preference.type :
        raise TypeError(f"LandmarkType {preference.type} does not match the type of Landmark {L[0].name}")

    for elem in L :
        elem.attractiveness = int(elem.attractiveness) + elem.n_tags*100        # arbitrary correction of the balance score vs number of tags
        elem.attractiveness = int(elem.attractiveness*preference.score/500)     # arbitrary computation

# Correct the score of a list of landmarks by taking into account preferences and the number of tags
def correct_score_test(L: List[Landmark], preference: Preference) :

    if len(L) == 0 :
        return
    
    if L[0].type != preference.type :
        raise TypeError(f"LandmarkType {preference.type} does not match the type of Landmark {L[0].name}")

    for elem in L :
        elem.attractiveness = int(elem.attractiveness/100) + elem.n_tags      # arbitrary correction of the balance score vs number of tags
        elem.attractiveness = elem.attractiveness*preference.score        # arbitrary computation

# Function to count elements within a 25m radius of a location
def count_elements_within_radius(coordinates: Tuple[float, float]) -> int:
    
    lat = coordinates[0]
    lon = coordinates[1]

    alpha = (180*RADIUS_CLOSE_TO)/(6371000*m.pi)

    bbox = {'latLower':lat-alpha,'lonLower':lon-alpha,'latHigher':lat+alpha,'lonHigher': lon+alpha}
    overpass = Overpass()
    
    # Build the query to find elements within the radius
    radius_query = overpassQueryBuilder(bbox=[bbox['latLower'],bbox['lonLower'],bbox['latHigher'],bbox['lonHigher']],
                             elementType=['node', 'way', 'relation'])

    try : 
        radius_result = overpass.query(radius_query)
    
        # The count is the number of elements found
        return radius_result.countElements()
    
    except :
        return None

# Creates a bounding box around precise coordinates
def create_bbox(coordinates: Tuple[float, float], side_length: int) -> Tuple[float, float, float, float]:
    """
    Create a simple bounding box around given coordinates.
    :param coordinates: tuple (lat, lon)
    -> lat: Latitude of the center point.
    -> lon: Longitude of the center point.
    :param side_length: int     - side length of the bbox in km
    :return: Bounding box as (min_lat, min_lon, max_lat, max_lon).
    """
    lat = coordinates[0]
    lon = coordinates[1]

    # Half the side length in km (since it's a square bbox)
    half_side_length_km = side_length / 2.0

    # Convert distance to degrees
    lat_diff = half_side_length_km / 111  # 1 degree latitude is approximately 111 km
    lon_diff = half_side_length_km / (111 * m.cos(m.radians(lat)))  # Adjust for longitude based on latitude

    # Calculate bbox
    min_lat = lat - lat_diff
    max_lat = lat + lat_diff
    min_lon = lon - lon_diff
    max_lon = lon + lon_diff

    return min_lat, min_lon, max_lat, max_lon

# Generates the list of landmarks for a given Landmarktype. Needs coordinates, a list of amenities and the corresponding LandmarkType
def get_landmarks_coords(coordinates: Tuple[float, float], l: List[Landmark], landmarktype: LandmarkType)->List[Landmark]:

    overpass = Overpass()

    # Generate a bbox around current coordinates
    bbox = create_bbox(coordinates, BBOX_SIDE)

    # Initialize some variables
    N = 0
    L = []

    for amenity in l :
        query = overpassQueryBuilder(bbox=bbox, elementType=['way', 'relation'], selector=amenity, includeCenter=True, out='body')
        result = overpass.query(query)
        N += result.countElements()

        for elem in result.elements():

            name = elem.tag('name')                             # Add name, decode to ASCII
            location = (elem.centerLat(), elem.centerLon())     # Add coordinates (lat, lon)

            # skip if unprecise location
            if name is None or location[0] is None:
                continue

            # skip if unused
            if 'disused:leisure' in elem.tags().keys():
                    continue
            
            else :
                
                osm_type = elem.type()              # Add type : 'way' or 'relation'
                osm_id = elem.id()                  # Add OSM id 
                elem_type = landmarktype            # Add the landmark type as 'sightseeing
                n_tags = len(elem.tags().keys())    # Add number of tags
                
                # Add score of given landmark based on the number of surrounding elements. Penalty for churches as there are A LOT
                if amenity == "'amenity'='place_of_worship'" :
                    score = int(count_elements_within_radius(location)*CHURCH_PENALTY)  
                elif amenity == "'leisure'='park'" :
                    score = int(count_elements_within_radius(location)*PARK_COEFF)  
                else :
                    score = count_elements_within_radius(location)

                if name == "Jardin du Luxembourg" :
                    continue

                if score is not None :
                    # Generate the landmark and append it to the list
                    landmark = Landmark(name=name, type=elem_type, location=location, osm_type=osm_type, osm_id=osm_id, attractiveness=score, must_do=False, n_tags=n_tags)
                    L.append(landmark)

    return L

def get_landmarks_nominatim(city_country: str, l: List[Landmark], landmarktype: LandmarkType)->List[Landmark] :
    
    overpass = Overpass()
    nominatim = Nominatim()
    areaId = nominatim.query(city_country).areaId()
    
    # Initialize some variables
    N = 0
    L = []

    for amenity in l :
        query = overpassQueryBuilder(area=areaId, elementType=['way', 'relation'], selector=amenity, includeCenter=True, out='body')
        result = overpass.query(query)
        N += result.countElements()

        for elem in result.elements():

            name = elem.tag('name')                             # Add name
            location = (elem.centerLat(), elem.centerLon())     # Add coordinates (lat, lon)

            # skip if unprecise location
            if name is None or location[0] is None:
                continue

            # skip if unused
            if 'disused:leisure' in elem.tags().keys():
                    continue
            
            else :
                
                osm_type = elem.type()              # Add type : 'way' or 'relation'
                osm_id = elem.id()                  # Add OSM id 
                elem_type = landmarktype            # Add the landmark type as 'sightseeing
                n_tags = len(elem.tags().keys())    # Add number of tags
                
                # Add score of given landmark based on the number of surrounding elements
                if amenity == "'amenity'='place_of_worship'" :
                    score = int(count_elements_within_radius(location)*CHURCH_PENALTY)
                else :
                    score = count_elements_within_radius(location)


                if score is not None :
                    # Generate the landmark and append it to the list
                    landmark = Landmark(name=name, type=elem_type, location=location, osm_type=osm_type, osm_id=osm_id, attractiveness=score, must_do=False, n_tags=n_tags)
                    L.append(landmark)

    return L
