import numpy as np

from typing import List
from typing import Tuple
from scipy.optimize import linprog
from scipy.linalg import block_diag
from structs.landmarks import Landmark
from math import radians, sin, cos, acos



DETOUR_FACTOR = 1.3         # detour factor for straightline distance
AVG_WALKING_SPEED = 4.8     # average walking speed in km/h


# Function that returns the distance in meters from one location to another
def get_distance(p1: Tuple[float, float], p2: Tuple[float, float]) :
    
    # Compute the straight-line distance in km
    if p1 == p2 :
        return 0, 0
    else: 
        dist = 6371.01 * acos(sin(radians(p1[0]))*sin(radians(p2[0])) + cos(radians(p1[0]))*cos(radians(p2[0]))*cos(radians(p1[1]) - radians(p2[1])))

    # Consider the detour factor for average city
    wdist = dist*DETOUR_FACTOR

    # Time to walk this distance (in minutes)
    wtime = wdist/AVG_WALKING_SPEED*60

    if wtime > 15 :
        wtime = 5*round(wtime/5)
    else :
        wtime = round(wtime)
  

    return round(wdist, 1), wtime


# landmarks = [Landmark_1, Landmark_2, ...]

# Convert the solution of the optimization into the list of edges to follow. Order is taken into account
def untangle(resx: list) -> list:
    N = len(resx)                   # length of res
    L = int(np.sqrt(N))             # number of landmarks. CAST INTO INT but should not be a problem because N = L**2 by def.
    n_edges = resx.sum()      # number of edges

    order = []
    nonzeroind = np.nonzero(resx)[0] # the return is a little funny so I use the [0]

    nonzero_tup = np.unravel_index(nonzeroind, (L,L))

    indx = nonzero_tup[0].tolist()
    indy = nonzero_tup[1].tolist()

    vert = (indx[0], indy[0])

    order.append(vert[0])
    order.append(vert[1])
    
    while len(order) < n_edges + 1 :
        ind = indx.index(vert[1])

        vert = (indx[ind], indy[ind])

        order.append(vert[1])

    return order

    
# Just to print the result
def print_res(res, order, landmarks: List[Landmark], P) -> list:

    things = []

    """N = int(np.sqrt(len(X)))
    for i in range(N):
        print(X[i*N:i*N+N])
    print("Optimal value:", -res.fun)  # Minimization, so we negate to get the maximum
    print("Optimal point:", res.x)
    for i,x in enumerate(X) : X[i] = round(x,0)
    print(order)"""

    if len(order) == len(landmarks): 
        print('\nAll landmarks can be visited within max_steps, the following order is suggested : ')
    else :
        print('Could not visit all the landmarks, the following order is suggested : ')

    for idx in order : 
        print('- ' + landmarks[idx].name)
        things.append(landmarks[idx].name)

    steps = path_length(P, abs(res.x))
    print("\nMinutes walked : " + str(steps))
    print(f"\nVisited {len(order)} out of {len(landmarks)} landmarks")

    return things



# prevent the creation of similar circles
def prevent_circle(resx, landmarks: List[Landmark], A_ub, b_ub) -> bool:
    for i, elem in enumerate(resx):
        resx[i] = round(elem)
    
    N = len(resx)               # length of res
    L = int(np.sqrt(N))         # number of landmarks. CAST INTO INT but should not be a problem because N = L**2 by def.
    n_edges = resx.sum()        # number of edges

    
    nonzeroind = np.nonzero(resx)[0] # the return is a little funny so I use the [0]

    nonzero_tup = np.unravel_index(nonzeroind, (L,L))

    ind_a = nonzero_tup[0].tolist()
    vertices_visited = ind_a
    vertices_visited.remove(0)

    ones = [1]*L
    h = [0]*N
    for i in range(L) :
        if i in vertices_visited :
            h[i*L:i*L+L] = ones

    A_ub = np.vstack((A_ub, h))
    b_ub.append(len(vertices_visited)-1)

    return A_ub, b_ub


def break_circle2(circle_vertices, landmarks: List[Landmark], A_ub, b_ub) -> bool:
    
    L = len(landmarks)         # number of landmarks. CAST INTO INT but should not be a problem because N = L**2 by def.


    if L-1 in circle_vertices :
        circle_vertices.remove(L-1)

    ones = [1]*L
    h = [0]*L*L
    for i in range(L) :
        if i in circle_vertices :
            h[i*L:i*L+L] = ones

    A_ub = np.vstack((A_ub, h))
    b_ub.append(len(circle_vertices)-1)

    return A_ub, b_ub


# Checks if the path is connected
def is_connected(resx, landmarks: List[Landmark]) -> bool:
    for i, elem in enumerate(resx):
        resx[i] = round(elem)
    
    N = len(resx)               # length of res
    L = int(np.sqrt(N))         # number of landmarks. CAST INTO INT but should not be a problem because N = L**2 by def.
    n_edges = resx.sum()        # number of edges

    
    nonzeroind = np.nonzero(resx)[0] # the return is a little funny so I use the [0]

    nonzero_tup = np.unravel_index(nonzeroind, (L,L))

    ind_a = nonzero_tup[0].tolist()
    ind_b = nonzero_tup[1].tolist()

    edges = []
    edges_visited = []
    vertices_visited = []

    edge1 = (ind_a[0], ind_b[0])
    edges_visited.append(edge1)
    vertices_visited.append(edge1[0])

    for i, a in enumerate(ind_a) :
        edges.append((a, ind_b[i]))      # Create the list of edges

    flag = False

    remaining = edges
    remaining.remove(edge1)
    # This can be further optimized
    #while len(vertices_visited) < n_edges + 1 :
    break_flag = False
    while len(remaining) > 0 and not break_flag:
        for edge2 in remaining :
            if edge2[0] == edge1[1] :
                if edge1[1] in vertices_visited :
                    edges_visited.append(edge2)
                    break_flag = True
                    break
                    #continue # continue vs break vs needed at all ?
                else :    
                    vertices_visited.append(edge1[1])
                    edges_visited.append(edge2)
                    remaining.remove(edge2)
                    edge1 = edge2

            elif edge1[1] == L-1 or edge1[1] in vertices_visited:
                        break_flag = True
                        break
                #break
        #if flag is True :
        #    break
    vertices_visited.append(edge1[1])


    if len(vertices_visited) == n_edges +1 :
        flag = True
        circle = []
    else: 
        flag = False
        circle = edges_visited

    """j = 0
    for i in vertices_visited :
        if landmarks[i].name == 'start' :
            ordered_visit = vertices_visited[j:] + vertices_visited[:j]
            break
        j+=1"""


    return flag, vertices_visited, circle




# Checks for cases of circular symmetry in the result
def has_circle(resx: list) :
    N = len(resx)               # length of res
    L = int(np.sqrt(N))         # number of landmarks. CAST INTO INT but should not be a problem because N = L**2 by def.
    n_edges = resx.sum()        # number of edges

    
    nonzeroind = np.nonzero(resx)[0] # the return is a little funny so I use the [0]

    nonzero_tup = np.unravel_index(nonzeroind, (L,L))

    indx = nonzero_tup[0].tolist()
    indy = nonzero_tup[1].tolist()
    

    verts = []

    for i, x in enumerate(indx) :
        verts.append((x, indy[i]))

    
    for vert in verts :
        visited = []
        visited.append(vert)

        while len(visited) < n_edges + 1 :

            try : 
                ind = indx.index(vert[1])

                vert = (indx[ind], indy[ind])

                if vert in visited :
                    return visited
                else :
                    visited.append(vert)
            except :
                break

    return []

# Constraint to not have d14 and d41 simultaneously. Does not prevent circular symmetry with more elements
def break_sym(N, A_ub, b_ub):
    upper_ind = np.triu_indices(N,0,N)

    up_ind_x = upper_ind[0]
    up_ind_y = upper_ind[1]

    for i, _ in enumerate(up_ind_x) :
        l = [0]*N*N
        if up_ind_x[i] != up_ind_y[i] :
            l[up_ind_x[i]*N + up_ind_y[i]] = 1
            l[up_ind_y[i]*N + up_ind_x[i]] = 1

            A_ub = np.vstack((A_ub,l))
            b_ub.append(1)

            """for i in range(7):
                print(l[i*7:i*7+7])
            print("\n")"""

    return A_ub, b_ub

# Constraint to not have circular paths. Want to go from start -> finish without unconnected loops
def break_circle(L, A_ub, b_ub, circle) :
    l = [0]*L*L

    for index in circle :
        x = index[0]
        y = index[1]
        l[x*L+y] = 1

    A_ub = np.vstack((A_ub,l))
    b_ub.append(len(circle)-1)

    """print("\n\nPREVENT CIRCLE")
    for i in range(7):
        print(l[i*7:i*7+7])
    print("\n")"""

    return A_ub, b_ub

# Constraint to respect max number of travels
def respect_number(N, A_ub, b_ub):
    """h = []
    for i in range(N) : h.append([1]*N)
    T = block_diag(*h)
    for l in T :
        for i in range(7):
            print(l[i*7:i*7+7])
        print("\n")"""
    #return np.vstack((A_ub, T)), b_ub + [1]*N
    ones = [1]*N
    zeros = [0]*N
    for i in range(N) :
        h = zeros*i + ones + zeros*(N-1-i)

        A_ub = np.vstack((A_ub, h))
        b_ub.append(1)

    return A_ub, b_ub
        

# Constraint to tie the problem together. Necessary but not sufficient to avoid circles
def respect_order(N: int, A_eq, b_eq): 
    for i in range(N-1) :     # Prevent stacked ones
        if i == 0 or i == N-1:  # Don't touch start or finish
            continue
        else : 
            l = [0]*N
            l[i] = -1
            l = l*N
            for j in range(N) :
                l[i*N + j] = 1

            A_eq = np.vstack((A_eq,l))
            b_eq.append(0)

    return A_eq, b_eq

# Compute manhattan distance between 2 locations
def manhattan_distance(loc1: tuple, loc2: tuple):
    x1, y1 = loc1
    x2, y2 = loc2
    return abs(x1 - x2) + abs(y1 - y2)

# Constraint to not stay in position. Removes d11, d22, d33, etc.
def init_eq_not_stay(N: int): 
    l = [0]*N*N

    for i in range(N) :
        for j in range(N) :
            if j == i :
                l[j + i*N] = 1
    
    l = np.array(np.array(l))

    return [l], [0]

# Constraint to ensure start at start and finish at goal
def respect_start_finish(N, A_eq: list, b_eq: list):
    ls = [1]*N + [0]*N*(N-1)    # sets only horizontal ones for start (go from)
    ljump = [0]*N*N
    ljump[N-1] = 1              # Prevent start finish jump
    lg = [0]*N*N
    for k in range(N-1) :       # sets only vertical ones for goal (go to)
        if k != 0 :             # Prevent the shortcut start -> finish
            lg[k*N+N-1] = 1 

    A_eq = np.vstack((A_eq,ls))
    A_eq = np.vstack((A_eq,ljump))
    A_eq = np.vstack((A_eq,lg))
    b_eq.append(1)
    b_eq.append(0)
    b_eq.append(1)

    return A_eq, b_eq

# Initialize A and c. Compute the distances from all landmarks to each other and store attractiveness
# We want to maximize the sightseeing :  max(c) st. A*x < b   and   A_eq*x = b_eq
def init_ub_dist(landmarks: List[Landmark], max_steps: int):
    # Objective function coefficients. a*x1 + b*x2 + c*x3 + ...
    c = []
    # Coefficients of inequality constraints (left-hand side)
    A = []
    for i, spot1 in enumerate(landmarks) :
        dist_table = [0]*len(landmarks)
        c.append(-spot1.attractiveness)
        for j, spot2 in enumerate(landmarks) :
            d, t = get_distance(spot1.location, spot2.location)
            dist_table[j] = t
        A.append(dist_table)
    c = c*len(landmarks)
    A_ub = []
    for line in A :
        #print(line)
        A_ub += line
    return c, A_ub, [max_steps]

# Go through the landmarks and force the optimizer to use landmarks where attractiveness is set to -1
def respect_user_mustsee(landmarks: List[Landmark], A_eq: list, b_eq: list) :
    L = len(landmarks)
    H = 0       # sort of heuristic to get an idea of the number of steps needed
    
    elem_prev = landmarks[0]
    
    for i, elem in enumerate(landmarks) :
        if elem.must_do is True and elem.name not in ['finish', 'start']:
            l = [0]*L*L
            for j in range(L) :     # sets the horizontal ones (go from)
                l[j +i*L] = 1       # sets the vertical ones (go to)        double check if good
                
            for k in range(L-1) :
                l[k*L+L-1] = 1  


            A_eq = np.vstack((A_eq,l))
            b_eq.append(2)

        d, t = get_distance(elem.location, elem_prev.location)
        H += t
        elem_prev = elem

            

    return A_eq, b_eq, H

# Computes the path length given path matrix (dist_table) and a result
def path_length(P: list, resx: list) :
    return np.dot(P, resx)

# Main optimization pipeline
def solve_optimization (landmarks, max_steps, printing_details) :

    N = len(landmarks)

    # SET CONSTRAINTS FOR INEQUALITY
    c, A_ub, b_ub = init_ub_dist(landmarks, max_steps)      # Add the distances from each landmark to the other
    P = A_ub                                                # store the paths for later. Needed to compute path length
    A_ub, b_ub = respect_number(N, A_ub, b_ub)              # Respect max number of visits (no more possible stops than landmarks). 

    # TODO : Problems with circular symmetry
    A_ub, b_ub = break_sym(N, A_ub, b_ub)                  # break the symmetry. Only use the upper diagonal values

    # SET CONSTRAINTS FOR EQUALITY
    A_eq, b_eq = init_eq_not_stay(N)                                # Force solution not to stay in same place
    A_eq, b_eq, H = respect_user_mustsee(landmarks, A_eq, b_eq)     # Check if there are user_defined must_see. Also takes care of start/goal
    A_eq, b_eq = respect_start_finish(N, A_eq, b_eq)                # Force start and finish positions
    A_eq, b_eq = respect_order(N, A_eq, b_eq)                       # Respect order of visit (only works when max_steps is limiting factor)

    # Bounds for variables (x can only be 0 or 1)
    x_bounds = [(0, 1)] * len(c)

    # Solve linear programming problem

    res = linprog(c, A_ub=A_ub, b_ub=b_ub, A_eq=A_eq, b_eq = b_eq, bounds=x_bounds, method='highs', integrality=3)
    


    # Raise error if no solution is found
    if not res.success :

        # Override the max_steps using the heuristic
        for i, val in enumerate(b_ub) :
            if val == max_steps : b_ub[i] = H

        # Solve problem again :
        res = linprog(c, A_ub=A_ub, b_ub=b_ub, A_eq=A_eq, b_eq = b_eq, bounds=x_bounds, method='highs', integrality=3)

        if not res.success :
            s = "No solution could be found, even when increasing max_steps using the heuristic"
            return s
            #raise ValueError("No solution could be found, even when increasing max_steps using the heuristic")
    
    # If there is a solution, we're good to go, just check for
    else :
        t, order, circle = is_connected(res.x, landmarks)
        i = 0

        # Break the circular symmetry if needed
        while len(circle) != 0 :
            A_ub, b_ub = prevent_circle(res.x, landmarks, A_ub, b_ub)
            A_ub, b_ub = break_circle(len(landmarks), A_ub, b_ub, circle)
            A_ub, b_ub = break_circle2(order, landmarks, A_ub, b_ub)
            res = linprog(c, A_ub=A_ub, b_ub=b_ub, A_eq=A_eq, b_eq = b_eq, bounds=x_bounds, method='highs', integrality=3)
            t, order, circle = is_connected(res.x, landmarks)
            if t :
                break
            #circle = has_circle(res.x)
            print(i)
            i += 1

        t, order, [] = is_connected(res.x, landmarks)
        
        if printing_details is True :
            if i != 0 :
                print(f"Neded to recompute paths {i} times because of unconnected loops...")
            
            print_res(res, order, landmarks, P)

        return order






